﻿using Carmenta.Engine;
using CarmentaWpfClient.Models;
using CarmentaWpfClient.ViewModels;
using CarmentaWpfClient.Utils;

using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using System;
using System.Windows;
using System.Reflection;
using System.IO;

namespace CarmentaWpfClient
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App : Application
    {
        private readonly IHost _host;

        public App()
        {
            Runtime.Initialize();

            _host = new HostBuilder()
                .ConfigureServices((_, services) => ConfigureServices(services))
                .Build();

            Services = _host.Services;
        }

        public new static App Current => (App)Application.Current;

        public IServiceProvider Services { get; }

        private static ServiceProvider ConfigureServices(IServiceCollection services)
        {
            services.AddSingleton<MapProvider>(x =>
            {
                string? pxFileDirectoryPath = Path.GetDirectoryName(
                    Assembly.GetExecutingAssembly().Location);
                return new MapProvider(pxFileDirectoryPath);
            });

            services.AddTransient<MapViewModel>();
            services.AddTransient<MainViewModel>();
            services.AddTransient<DrawerMenuViewModel>();
            services.AddTransient<LayerControlViewModel>();
            services.AddTransient<ToolsMenuViewModel>();
            services.AddTransient<TacticalSymbolsMenuViewModel>();
            services.AddTransient<IFeatureSerializer, BinaryFeatureSerializer>();

            services.AddHostedService<StreamCommunicator>(x =>
            {
                if (!x.GetService<MapProvider>().TryGetMap("View0", out MapModel mapModel))
                    throw new Exception("Could not load map configuration.");

                MemoryDataSet memoryDataSet = mapModel.GetMemoryDataSet("TacticalSymbolsDataSet") 
                                              ?? throw new Exception("Could not find TacticalSymbolsDataSet.");

                IFeatureSerializer serializer = x.GetService<IFeatureSerializer>()
                                                ?? throw new Exception("Could not get feature serializer implementation.");

                return new StreamCommunicator(memoryDataSet, serializer);
            });

            return services.BuildServiceProvider();
        }

        protected override async void OnStartup(StartupEventArgs e)
        {
            await _host.StartAsync();
            base.OnStartup(e);
        }

        protected override async void OnExit(ExitEventArgs e)
        {
            using (_host)
            {
                await _host.StopAsync();
            }
            base.OnExit(e);
        }
    }
}