﻿using Carmenta.Engine;
using CarmentaWpfClient.Messages;
using CommunityToolkit.Mvvm.Messaging;
using Microsoft.AspNetCore.SignalR.Client;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace CarmentaWpfClient.Utils;

public class StreamCommunicator : BackgroundService
{
    private readonly MemoryDataSet _memoryDataSet;
    private readonly IFeatureSerializer _featureSerializer;
    private readonly HubConnection _hubConnection;

    private readonly ConcurrentDictionary<Guid, Id> _idMap = new();
    private readonly ConcurrentDictionary<Id, Guid> _reverseIdMap = new();

    public StreamCommunicator(MemoryDataSet dataSet, IFeatureSerializer featureSerializer)
    {
        _featureSerializer = featureSerializer;
        _memoryDataSet = dataSet;

        _memoryDataSet.EnableEvents = true;
        _memoryDataSet.FeatureInserted += FeatureInserted;
        _memoryDataSet.FeatureChanged += FeatureChanged;
        _memoryDataSet.FeatureRemoved += FeatureRemoved;

        _hubConnection = new HubConnectionBuilder()
            .WithUrl("https://localhost:5001/communicationHub")
            .WithAutomaticReconnect()
            .AddMessagePackProtocol()
            .Build();

        _hubConnection.Closed += OnHubConnectionClosed;
        _hubConnection.Reconnected += OnHubReconnected;
    }

    private Task OnHubReconnected(string? arg)
    {
        IsConnected = true;
        return Task.CompletedTask;
    }

    private Task OnHubConnectionClosed(Exception? _)
    {
        IsConnected = false;

        return Task.CompletedTask;
    }

    public bool IsConnected { get; private set; }

    public async Task InitializeAsync()
    {
        try
        {
            await _hubConnection.StartAsync();

            var existingFeatures = await _hubConnection.InvokeAsync<List<(Guid, byte[])>>("AllFeatures");

            foreach ((Guid id, byte[] buffer) in existingFeatures)
                HandleStreamAdded(id, buffer);

            await SetupStreamListeners();
        }
        catch(Exception)
        {
            //Log error etc..
        }
    }

    private Task SetupStreamListeners()
    {
        Task addedStreamTask = SetupAddedStream();
        Task editStreamTask = SetupStreamEdited();
        Task removeStreamTask = SetupRemoveStream();

        return Task.WhenAll(addedStreamTask, editStreamTask, removeStreamTask);
    }

    private async Task SetupAddedStream()
    {
        IAsyncEnumerable<(Guid Id, byte[] Buffer)> addedStream = _hubConnection.StreamAsync<(Guid Id, byte[] Buffer)>("AddedStream");

        await foreach ((Guid id, byte[] buffer) in addedStream)
            HandleStreamAdded(id, buffer);
    }

    private void HandleStreamAdded(Guid id, byte[] buffer)
    {
        if (_idMap.ContainsKey(id))
            return;

        using var guard = new Guard(_memoryDataSet);

        Feature feature = _featureSerializer.Deserialize(buffer);

        // Temporarily turn off events to avoid triggering another
        // inserted event which could cause recursive behaviour.
        _memoryDataSet.EnableEvents = false;
        Id featureId = _memoryDataSet.Insert(feature);
        _memoryDataSet.EnableEvents = true;

        _idMap[id] = featureId;
        _reverseIdMap[featureId] = id;

        WeakReferenceMessenger.Default.Send(new UpdateViewMessage());
    }

    private async Task SetupStreamEdited()
    {
        IAsyncEnumerable<(Guid Id, byte[] Buffer)> editedStream = _hubConnection.StreamAsync<(Guid Id, byte[] Buffer)>("UpdateStream");

        await foreach((Guid id, byte[] buffer) in editedStream)
            HandleStreamEdited(id, buffer);
    }

    private void HandleStreamEdited(Guid id, byte[] buffer)
    {
        if (!_idMap.ContainsKey(id))
        {
            HandleStreamAdded(id, buffer);
            return;
        }

        Feature feature = _featureSerializer.Deserialize(buffer);
        Id existingFeatureId = _idMap[id];

        // Optimize by replacing with dictionary lookup
        Feature? existingFeature = _memoryDataSet.GetFeature(existingFeatureId);

        UpdateFeature(existingFeature, feature.Geometry, feature.Attributes);
    }

    private async Task SetupRemoveStream()
    {
        IAsyncEnumerable<Guid> removeStream = _hubConnection.StreamAsync<Guid>("RemoveStream");

        await foreach(Guid id in removeStream)
        {
            if (!_idMap.ContainsKey(id))
                continue;

            _idMap.Remove(id, out Id idToRemove);
            _reverseIdMap.Remove(idToRemove, out _);
            using var guard = new Guard(_memoryDataSet);

            _memoryDataSet.EnableEvents = false;
            _memoryDataSet.Remove(idToRemove);
            _memoryDataSet.EnableEvents = true;
        }
    }

    private void UpdateFeature(Feature feature, Geometry geometry, AttributeSet attributes)
    {
        switch (geometry)
        {
            case PointGeometry point:
                feature.GetGeometryAsPoint().Point = point.Point;
                break;
            case LineGeometry line:
                {
                    LineGeometry? geom = feature.GetGeometryAsLine();
                    geom.Points.Clear();
                    geom.Points.AddRange(line.Points);
                    break;
                }
            case PolygonGeometry polygon:
                {
                    PolygonGeometry? geom = feature.GetGeometryAsPolygon();
                    geom.Rings.Clear();
                    geom.Rings.AddRange(polygon.Rings.Clone() as LineGeometryCollection);
                    break;
                }
            default:
                return;
        }

        feature.Attributes.Merge(attributes.Clone() as AttributeSet);

        _memoryDataSet.RefreshFeaturePresentation(feature, false);
        WeakReferenceMessenger.Default.Send(new UpdateViewMessage());
    }

    private async void FeatureChanged(object sender, MemoryDataSetEventArgs args)
    {
        Guid corrId = _reverseIdMap[args.Feature.Id];

        byte[] buffer = _featureSerializer.Serialize(args.Feature);
        await _hubConnection.InvokeAsync("UpdateFeature", corrId, buffer);
    }

    private async void FeatureInserted(object sender, MemoryDataSetEventArgs args)
    {
        Feature? feature = args.Feature;
        var corrId = Guid.NewGuid();

        _idMap[corrId] = feature.Id;
        _reverseIdMap[feature.Id] = corrId;

        byte[] buffer = _featureSerializer.Serialize(feature);
        await _hubConnection.InvokeAsync("AddFeature", corrId, buffer);
        WeakReferenceMessenger.Default.Send(new UpdateViewMessage());
    }

    private async void FeatureRemoved(object sender, MemoryDataSetEventArgs args)
    {
        Guid corrId = _reverseIdMap[args.Feature.Id];
        await _hubConnection.InvokeAsync("RemoveFeature", corrId);

        WeakReferenceMessenger.Default.Send(new UpdateViewMessage());
    }

    public override async Task StopAsync(CancellationToken cancellationToken)
    {
        if (_hubConnection.State == HubConnectionState.Connected)
            await _hubConnection.StopAsync(cancellationToken);
    }

    protected override async Task ExecuteAsync(CancellationToken stoppingToken)
    {
        await InitializeAsync();
    }
}
