﻿using Carmenta.Engine;

using CarmentaWpfClient.Messages;
using CarmentaWpfClient.Models;
using CarmentaWpfClient.Utils;

using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Messaging;

using System;
using System.Collections.Generic;
using System.Linq;

namespace CarmentaWpfClient.ViewModels
{
    public partial class MapViewModel : ObservableObject,
        IRecipient<UpdateViewMessage>, IRecipient<ChangeToolMessage>,
        IRecipient<SymbolSelectedMessage>, IRecipient<DeleteSelectedFeatures>
    {
        private CreateTool? _createTool;

        private CreateTool? _tacticalCreateTool;

        private StandardTool _standardTool = new();

        private readonly Dictionary<ulong, MemoryDataSet> _dataSetLookup;

        private IdCollection _selectedIds = new();

        [ObservableProperty]
        private ITool? _currentTool;

        private readonly MapProvider _mapProvider;

        [ObservableProperty]
        private MapModel? _currentMap;

        public MapViewModel(MapProvider mapProvider)
        {
            WeakReferenceMessenger.Default.RegisterAll(this);
            _mapProvider = mapProvider;

            CurrentMap = _mapProvider.Maps.First();
            CurrentTool = _standardTool;

            var featureMemoryDataSet = CurrentMap.View
                .GetTypedObject<MemoryDataSet>("FeatureMemoryDataSet");

            if (featureMemoryDataSet is not null)
                _createTool = new CreateTool(featureMemoryDataSet);

            var tacticalSymbolsDataSet = CurrentMap.View
                .GetTypedObject<MemoryDataSet>("TacticalSymbolsDataSet");

            if (tacticalSymbolsDataSet is not null)
                _tacticalCreateTool = new CreateTool(tacticalSymbolsDataSet);

            _dataSetLookup = new Dictionary<ulong, MemoryDataSet>
            {
                {featureMemoryDataSet.Id, featureMemoryDataSet },
                {tacticalSymbolsDataSet.Id, tacticalSymbolsDataSet }
            };
            CurrentMap.View.SelectionChanged += ViewSelectionChanged;
        }

        private void ViewSelectionChanged(object? _, EventArgs __)
        {
            _selectedIds = CurrentMap.View.GetSelectedIds();
            OnPropertyChanged(nameof(CanDeleteFeatures));
        }

        public bool CanDeleteFeatures => _selectedIds.Count > 0;

        public void Receive(UpdateViewMessage message) => CurrentMap?.Update();

        public void Receive(ChangeToolMessage message)
        {
            if (message is { ToolType: ToolType.Create, ToolCreateMode: not null })
            {
                if (_createTool == null)
                {
                    return;
                }

                _createTool.CreateMode = message.ToolCreateMode.Value;
                CurrentTool = _createTool;
            }
            else
                CurrentTool = _standardTool;
        }

        public void Receive(SymbolSelectedMessage message)
        {
            if (!message.SymbolMetadata.Supported)
                return;

            string sidc = message.SymbolMetadata.Sidc[..3] + GetAffiliation(message.Affiliation) + message.SymbolMetadata.Sidc[4..];

            try
            {
                var createToolParameters = new MilitaryDCreateToolParameters(sidc);
                _tacticalCreateTool.Parameters = createToolParameters;

                CurrentTool = _tacticalCreateTool;
            }
            catch (EngineException ex)
            {
                // Logging, etc..
            }
        }

        private static char GetAffiliation(string affiliation) => affiliation switch
        {
            "Pending" => '0',
            "Unknown" => '1',
            "Friend" => '3',
            "Neutral" => '4',
            "Hostile" => '6',
            _ => '1'
        };

        public void Receive(DeleteSelectedFeatures message)
        {
            foreach (var id in _selectedIds)
            {
                var ds = _dataSetLookup[id.DataSetId];
                using var guard = new Guard(ds);

                ds.Remove(id);
            }
        }
    }
}
