﻿using System.Collections.Concurrent;
using System.Reactive.Subjects;

namespace StreamingService;

public class StreamingRepository<TKey, TValue> where TKey : notnull
{
    private readonly ConcurrentDictionary<TKey, TValue> _store = new();

    private readonly Subject<(TKey key, TValue value)> _addedSubject = new();
    private readonly Subject<(TKey key, TValue value)> _updatedSubject = new();
    private readonly Subject<TKey> _removedSubject = new();

    public IObservable<(TKey key, TValue value)> AddedStream => _addedSubject;
    public IObservable<(TKey key, TValue value)> UpdateStream => _updatedSubject;
    public IObservable<TKey> RemovedStream => _removedSubject;

    public void Add(TKey key, TValue value)
    {
        _store[key] = value;
        _addedSubject.OnNext((key, value));
    }

    public void Update(TKey key, TValue value)
    {
        if (!_store.ContainsKey(key))
            return;

        _store[key] = value;
        _updatedSubject.OnNext((key, value));
    }

    public void Remove(TKey key)
    {
        if (_store.TryRemove(key, out _))
            _removedSubject.OnNext(key);
    }

    public IEnumerable<(TKey, TValue value)> Objects => _store.Select(x => (x.Key, x.Value));
}
